@@ -13,6 +13,13 @@ ppMethods <- c("BoxCox", "YeoJohnson", "expoTrans", "invHyperbolicSine",
 
 invHyperbolicSineFunc <- function(x) log(x+sqrt(x^2+1))
 
+getRangeBounds <- function(pp) {
+ if(!is.null(pp$rangeBounds)) {
+   list(lower = pp$rangeBounds[1], upper = pp$rangeBounds[2])
+ } else {
+   list(lower = 0.0, upper = 1.0)
+ }
+}
 
 #' Pre-Processing of Predictors
 #' 
@@ -41,7 +48,7 @@ invHyperbolicSineFunc <- function(x) log(x+sqrt(x^2+1))
 #' from the data in \code{x}) from the predictor values while \code{method =
 #' "scale"} divides by the standard deviation.
 #' 
-#' The "range" transformation scales the data to be within [0, 1]. If new
+#' The "range" transformation scales the data to be within \code{rangeBounds}. If new
 #' samples have values larger or smaller than those in the training set, values
 #' will be outside of this range.
 #' 
@@ -129,6 +136,8 @@ invHyperbolicSineFunc <- function(x) log(x+sqrt(x^2+1))
 #' the number of total samples. See \code{\link{nearZeroVar}}. 
 #' @param cutoff a numeric value for the pair-wise absolute correlation cutoff. 
 #' See \code{\link{findCorrelation}}.  
+#' @param rangeBounds a two-element numeric vector specifying closed interval
+#' for range transformation
 #' @param \dots additional arguments to pass to \code{\link[fastICA]{fastICA}},
 #' such as \code{n.comp}
 #' @return \code{preProcess} results in a list with elements \item{call}{the
@@ -150,7 +159,7 @@ invHyperbolicSineFunc <- function(x) log(x+sqrt(x^2+1))
 #' @seealso \code{\link{BoxCoxTrans}}, \code{\link{expoTrans}}
 #' \code{\link[MASS]{boxcox}}, \code{\link[stats]{prcomp}},
 #' \code{\link[fastICA]{fastICA}}, \code{\link{spatialSign}}
-#' @references \url{http://topepo.github.io/caret/preprocess.html}
+#' @references \url{http://topepo.github.io/caret/pre-processing.html}
 #' 
 #' Kuhn and Johnson (2013), Applied Predictive Modeling, Springer, New York
 #' (chapter 4)
@@ -187,7 +196,6 @@ preProcess <- function(x, ...) UseMethod("preProcess")
 
 #' @rdname preProcess
 #' @importFrom stats complete.cases median sd prcomp
-#' @importFrom car powerTransform yjPower
 #' @export
 preProcess.default <- function(x, method = c("center", "scale"),
                                thresh = 0.95,
@@ -202,12 +210,15 @@ preProcess.default <- function(x, method = c("center", "scale"),
                                freqCut = 95/5, 
                                uniqueCut = 10,
                                cutoff = 0.9,
+                               rangeBounds = c(0, 1),
                                ...) {
+  if(!inherits(x, "matrix") & !inherits(x, "data.frame"))
+    stop("Matrices or data frames are required for preprocessing", call. = FALSE)
   column_types <- get_types(x)
   tmp <- pre_process_options(method, column_types)
   method <- tmp$opts
   wildcards <- tmp$wildcards
-
+  
   ## the row.norm option in fastICA states: "logical value indicating whether rows
   ## of the data matrix X should be standardized beforehand." Basically, this means that
   ## we would center *and* scale before the ICA step, so let's adjust the "scale" method too
@@ -256,7 +267,7 @@ preProcess.default <- function(x, method = c("center", "scale"),
     }
     method$conditionalX <- NULL
   }    
-
+  
   ## check for highly correlated predictors
   if(any(names(method) == "corr")){
     cmat <- try(cor(x[, !(colnames(x) %in% method$ignore), drop = FALSE], 
@@ -268,12 +279,20 @@ preProcess.default <- function(x, method = c("center", "scale"),
         removed <- colnames(cmat)[high_corr]
         method$remove <- unique(c(method$remove, removed))
         if(verbose) cat(paste(" ", length(removed), "highly correlated predictors were removed.\n"))
-        x <- x[, !(colnames(x) %in% removed)]
       } else warning(paste("correlation matrix could not be computed:\n", cmat))
     }
     method$corr <- NULL
   }
   
+  x <- x[, !(colnames(x) %in% method$remove), drop = FALSE] 
+  method = sapply(names(method), function(u) 
+    if(u != 'remove'){
+      method[[u]][ which(( method[[u]] %in% colnames(x)))]
+    } else { 
+      method[[u]]
+    }, simplify = FALSE
+  )
+  
   if(any(names(method) == "invHyperbolicSine")) {
     if(verbose) cat(" applying invHyperbolicSine\n")
     for(i in method$invHyperbolicSine) x[,i] <- invHyperbolicSineFunc(x[,i])
@@ -292,29 +311,34 @@ preProcess.default <- function(x, method = c("center", "scale"),
     for(i in method$BoxCox) x[,i] <- predict(bc[[i]], x[,i])
   } else bc <- NULL
   
+  
   if(any(names(method) == "YeoJohnson")) {
-    yj <- group_yj(x[, method$YeoJohnson, drop = FALSE],
-                   numUnique = numUnique,
-                   verbose = verbose)
+    yj <- vapply(
+      x[, method$YeoJohnson, drop = FALSE],
+      recipes::estimate_yj,
+      c(lambda = 0),
+      limits = c(-3, 3), ## consistent with `car` defaults
+      nunique = numUnique
+    )
+    yj <- yj[!is.null(yj) & !is.na(yj)]
     if(length(yj) > 0) {
       if(verbose) cat(" applying them to training data\n")
       if(length(yj) != length(method$YeoJohnson)) {
         method$YeoJohnson <- names(yj)
       }
-      lam <- unlist(lapply(yj, function(x) if(class(x) == "powerTransform") x$lambda else NA))
-      lam <- lam[!is.na(lam)]
-      if(length(lam) > 0) {
-        for(i in seq(along = lam)) {
-          who <-  gsub("\\.Y1$", "", names(lam)[i])
-          x[,who] <- yjPower(x[,who], yj[[who]]$lambda)
+      # now apply to current data
+      if(length(yj) > 0) {
+        for(i in seq(along = yj)) {
+          who <- names(yj)[i]
+          x[,who] <- recipes::yj_trans(x[,who], yj[who])
         }
       }
     } else {
       if(verbose) cat(" all of the transformations failed\n")
       yj <- NULL
     }
-  } else yj <- NULL
-  
+  } else yj <- NULL  
+
   if(any(names(method) == "expoTrans")) {
     if(verbose) 
       cat("Estimating exponential transformations for", 
@@ -324,7 +348,7 @@ preProcess.default <- function(x, method = c("center", "scale"),
                    expoTrans.default, numUnique = numUnique)
     } else {
       et <- apply(x[, method$expoTrans, drop = FALSE], 2,
-                   expoTrans.default, numUnique = numUnique)
+                  expoTrans.default, numUnique = numUnique)
     }
     if(verbose) cat(" applying them to training data\n")
     omit_expo <- NULL
@@ -369,6 +393,12 @@ preProcess.default <- function(x, method = c("center", "scale"),
   
   if(any(names(method) == "range")) {
     if(verbose) cat("Calculating", length(method$range), "statistcs for scaling to a range\n")
+    ## check rangeBounds consistency
+    if(!is.numeric(rangeBounds) || length(rangeBounds) != 2)
+      stop("'rangeBounds' should be a two-element numeric vector")
+    if(rangeBounds[1] >= rangeBounds[2])
+      stop("'rangeBounds' interval is empty")
+
     ranges <- apply(x[, method$range, drop = FALSE], 
                     2, 
                     function(x) c(min(x, na.rm = na.remove), 
@@ -382,7 +412,9 @@ preProcess.default <- function(x, method = c("center", "scale"),
       method$range <- method$range[!(method$range %in% names(is_same)[is_same])]
     }
     x[, method$range] <- sweep(x[, method$range, drop = FALSE], 2, ranges[1,], "-")
-    x[, method$range] <- sweep(x[, method$range, drop = FALSE], 2, ranges[2,] - ranges[1,], "/")
+    x[, method$range] <- sweep(x[, method$range, drop = FALSE], 2,
+                               (ranges[2,] - ranges[1,]) / (rangeBounds[2] - rangeBounds[1]), "/")
+    x[, method$range] <- sweep(x[, method$range, drop = FALSE], 2, rangeBounds[1], "+")
   } else ranges <- NULL
   
   if(any(names(method) == "bagImpute")){
@@ -466,7 +498,8 @@ preProcess.default <- function(x, method = c("center", "scale"),
               bagImp = bagModels,
               median = medianValue,
               data = if(any(names(method) == "knnImpute")) 
-                x[complete.cases(x),method$knnImpute,drop = FALSE] else NULL)
+                x[complete.cases(x),method$knnImpute,drop = FALSE] else NULL,
+              rangeBounds = rangeBounds)
   out <- structure(out, class = "preProcess")
   out
 }
@@ -507,15 +540,15 @@ predict.preProcess <- function(object, newdata, ...) {
   }
   
   if(!is.null(object$yj)) {
-    lam <- unlist(lapply(object$yj, function(x) if(class(x) == "powerTransform") x$lambda else NA))
+    lam <- get_yj_lambda(object$yj)
     lam <- lam[!is.na(lam)]
     if(length(lam) > 0) {
       for(i in seq(along = lam)) {
-        who <-  gsub("\\.Y1$", "", names(lam)[i])
-        newdata[,who] <- yjPower(newdata[,who], object$yj[[who]]$lambda)
+        who <- names(lam)[i]
+        newdata[,who] <- recipes::yj_trans(newdata[,who], lam[who])
       }
     }
-  }  
+  } 
   
   if(!is.null(object$et)) {
     for(i in seq(along = object$et)) {
@@ -525,12 +558,17 @@ predict.preProcess <- function(object, newdata, ...) {
   }  
   
   if(any(names(object$method) == "range")) {
+    rangeBounds <- getRangeBounds(object)
     newdata[, object$method$range] <- 
       sweep(newdata[, object$method$range, drop = FALSE], 2, 
             object$ranges[1,], "-")
     newdata[, object$method$range] <- 
-      sweep(newdata[, object$method$range, drop = FALSE], 2, 
-            object$ranges[2,] - object$ranges[1,], "/")
+      sweep(newdata[, object$method$range, drop = FALSE], 2,
+            (object$ranges[2,] - object$ranges[1,]) /
+              (rangeBounds$upper - rangeBounds$lower), "/")
+    newdata[, object$method$range] <-
+      sweep(newdata[, object$method$range, drop = FALSE], 2,
+            rangeBounds$lower, "+")
   }
   
   if(any(names(object$method) == "center")) 
@@ -675,7 +713,10 @@ print.preProcess <- function(x, ...) {
   pp <- gsub("knnImpute", paste(x$k, "nearest neighbor imputation"), pp)
   pp <- gsub("bagImpute", "bagged tree imputation", pp)  
   pp <- gsub("medianImpute", "median imputation", pp)    
-  pp <- gsub("range", "re-scaling to [0, 1]", pp)  
+
+  rangeBounds <- getRangeBounds(x)
+  pp <- gsub("range", paste0("re-scaling to [", rangeBounds$lower, ", ", rangeBounds$upper, "]"), pp)
+
   pp <- gsub("remove", "removed", pp)  
   pp <- gsub("ignore", "ignored", pp)  
   
@@ -693,17 +734,18 @@ print.preProcess <- function(x, ...) {
     } else print(summary(unlist(lapply(x$bc, function(x) x$lambda))))
     cat("\n")
   }
+  
   if(any(names(x$method) == "YeoJohnson")) {
     cat("Lambda estimates for Yeo-Johnson transformation:\n")
-    if(length(x$yj) < 11) {
-      lmbda <- unlist(lapply(x$yj, function(x) if(class(x) == "powerTransform") x$lambda else NA))
+    lmbda <- get_yj_lambda(x$yj)
+    if(length(lmbda) < 11) {
       naLmbda <- sum(is.na(lmbda))
       cat(paste(round(lmbda[!is.na(lmbda)], 2), collapse = ", "))
       if(naLmbda > 0) cat(" (#NA: ", naLmbda, ")\n", sep = "")
-    } else print(summary(unlist(lapply(x$yj, function(x) if(class(x) == "powerTransform") x$lambda else NA))))
+    } else print(summary(lmbda))
     cat("\n")
   }  
-  
+
   if(any(names(x$method) == "pca")) {
     if(is.null(x$pcaComp)) {
       cat("PCA needed", x$numComp, ifelse(x$numComp > 1, "components", "component"),
@@ -867,7 +909,7 @@ pre_process_options <- function(opts, vars) {
   }
   
   if(any(methods %in% "range") & any(methods %in% c("center", "scale", "BoxCox")))
-    stop("Centering, scaling and/or Box-Cox transformations are inconsistent with scaling to a range of [0, 1]")  
+    stop("Centering, scaling and/or Box-Cox transformations are inconsistent with scaling to a range")
   
   ## coerce certain options based on others
   if("pca" %in% methods) {
@@ -1017,29 +1059,6 @@ group_bc <- function(x, outcome = NULL,
   bc[!is.null(bc) & !is.na(bc)]
 }
 
-yjWrap <- function(x, numUnique = numUnique) {
-  if(length(unique(x)) >= numUnique) {
-    out <- try(powerTransform(y ~ 1,
-                              data = data.frame(y = x[!is.na(x)]),
-                              family = "yjPower"),
-               silent = TRUE)
-    if(class(out)[1] == "try-error") out <- NA
-  } else out <- NA
-  out
-}
-
-group_yj <- function(x, numUnique, verbose) {
-  if(verbose) 
-    cat("Estimating Yeo-Johnson transformations for", ncol(x), "predictors...")
-  
-  if(is.matrix(x)) {
-    yj <- apply(x, 2, yjWrap, numUnique = numUnique)  
-  } else {
-    yj <- lapply(x, yjWrap, numUnique = numUnique)  
-  }   
-  yj[!is.null(yj) & !is.na(yj)]
-}
-
 convert_method <- function(x) {
   new_method <- list()
   if("center" %in% x$method)       new_method$center       <- names(x$mean)
@@ -1057,3 +1076,17 @@ convert_method <- function(x) {
   x$method <- new_method
   x
 }
+
+## code for using car method; extract lambdas either way and use
+## new code for predictions. Same for predict method
+get_yj_lambda <- function(x) {
+  if(inherits(x[[1]], "powerTransform")) {
+    # backwards compat with old caret objecgts that used `car`
+    res <- unlist(lapply(x, function(x) x$lambda))
+    names(res) <- gsub("\\.Y1$", "", names(res))
+  } else {
+    res <- x
+  }
+  res[!is.na(res)]
+}
+
